
#define TIME_STAMP_WRITE			0
#define TIME_STAMP_LAST_ACCESS		1
#define TIME_STAMP_CREATION			2

#define DATE_FORMAT_MDY				0
#define DATE_FORMAT_DMY				1
#define DATE_FORMAT_YMD				2

#define EPOCH						1601
#define ULTIMA						9999

#define WEEKDAYS_MASK				0x3e
#define WEEKENDS_MASK				0x41



const unsigned short int MonthLength[13][2] = { {NULL, NULL}, {31, 31}, {28, 29}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31} };
const unsigned int Kalends[13][2] = { {NULL, NULL}, {0, 0}, {31, 31}, {59, 60}, {90, 91}, {120, 121}, {151, 152}, {181, 182}, {212, 213}, {243, 244}, {273, 274}, {304, 305}, {334, 335} };

const wchar_t DowName[7][4] = { L"Sun", L"Mon", L"Tue", L"Wed", L"Thu", L"Fri", L"Sat" };



int FreeLinkedList( void **Pointer )
{
	if ( Pointer == NULL )
		return -1;

	size_t *This = (size_t*) *Pointer, Next = NULL;

	while ( This != NULL ) {
		// Printf( L"   * Freeing memory at 0x%08x\r\n", This );
		Next = *This;
		free( This );
		This = (size_t*) Next;
	}

	*Pointer = NULL;
	return 0;
}


size_t TrimString( LPTSTR String )
{
	if ( String == NULL )
		return -1;

	size_t j = 0, terminus = 0;
	wchar_t ch = 0x00;

	for ( size_t i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( ch == '"' )
			continue;

		if ( iswspace( ch ) ) {
			if ( j == 0 )
				continue;
		}
		else
			terminus = j + 1;
		
		String[j++] = ch;
	}

	String[terminus] = 0x00;

	return terminus;
}



BOOL IsColonOrEquals( wchar_t ch );

int DisplayErrorHeader();


struct _SizeRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	unsigned __int64 Smallest;
	unsigned __int64 Largest;
};

struct _TimeRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	unsigned short int WhichStamp;
	unsigned int EarliestTime;
	unsigned int LatestTime;
};

struct _DateRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	unsigned short int WhichStamp;
	unsigned int EarliestJulian;
	unsigned int LatestJulian;
	unsigned int EarliestTime;
	unsigned int LatestTime;
};

struct _ExclusionRange {
	void *Next;
	wchar_t Type;
	wchar_t Filespec;
};

struct _DOWRange {
	void *Next;
	wchar_t Type;
	unsigned short int WhichStamp;
	unsigned short int DOWMask;
};

struct _DescriptionRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	wchar_t Description;
};

struct _OwnerRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	wchar_t Owner;
};


int _Ranges_GetFileDescription( LPTSTR Filename, LPTSTR Description, unsigned int BufferLen )
{
	if ( !Filename || !Description )
		return -1;

	if ( Filename == Description )
		return -1;

	if ( Filename[0] == '\0' )
		return ERROR_FILE_NOT_FOUND;

	if ( !QueryIsFileOrDirectory( Filename ) )
		return ERROR_FILE_NOT_FOUND;


	wchar_t temp[MAX_ARG_LEN]		= L"%@descript[";


	wcscat_s( temp, MAX_ARG_LEN, Filename );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	ExpandVariables( temp, 1 );

	if ( (unsigned int) wcslen( temp ) >= BufferLen )
		return ERROR_BUFFER_OVERFLOW;

	wcscpy_s( Description, BufferLen, temp );
	return 0;
}


int _Ranges_AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}


int ParseSize2( wchar_t **ch, unsigned __int64 *Value )
{
	if ( ( ch == NULL ) || ( Value == NULL ) )
		return -1;
	else if ( iswdigit( **ch ) == 0 )
		return 1;

	*Value = 0;

	while ( iswdigit( **ch ) ) {
		*Value = ( *Value * 10 ) + ( **ch - '0' );
		(*ch)++;
	}

	switch ( **ch ) {
		case 'k' :
			*Value *= 1000LL;
			(*ch)++;
			break;

		case 'm' :
			*Value *= 1000000LL;
			(*ch)++;
			break;

		case 'g' :
			*Value *= 1000000000LL;
			(*ch)++;
			break;

		case 't' :
			*Value *= 1000000000000LL;
			(*ch)++;
			break;

		case 'p' :
			*Value *= 1000000000000000LL;
			(*ch)++;
			break;

		case 'K' :
			*Value = *Value << 10;
			(*ch)++;
			break;

		case 'M' :
			*Value = *Value << 20;
			(*ch)++;
			break;

		case 'G' :
			*Value = *Value << 30;
			(*ch)++;
			break;

		case 'T' :
			*Value = *Value << 40;
			(*ch)++;
			break;

		case 'P' :
			*Value = *Value << 50;
			(*ch)++;
			break;
	}

	return 0;
}


int ParseSizeRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	wchar_t *sep1 = inString;
	unsigned __int64 Smallest = 0, Largest = ULLONG_MAX;
	BOOL Add = FALSE;

	if ( IsColonOrEquals( *sep1 ) )
		sep1++;

	if ( wcscmp( sep1, L"]" ) == 0 )
		return 0;

	if ( iswdigit( *sep1 ) )
		ParseSize2( &sep1, &Smallest );

	if ( *sep1 == ',' )
		sep1++;

	if ( ( *sep1 == '+' ) && iswdigit( *(sep1 + 1) ) ) {
		Add = TRUE;
		sep1++;
	}

	if ( iswdigit( *sep1 ) )
		ParseSize2( &sep1, &Largest );

	if ( Add )
		Largest += Smallest;

	if ( Smallest > Largest ) {
		__int64 t = Smallest;
		Smallest = Largest;
		Largest = t;
	}

	if ( wcscmp( sep1, L"]" ) ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE), L"Bad size range: \"%s\"\r\n", inString );
		return 1;
	}

	// Printf( L"   * Size range = %q - %q %s\r\n", Smallest, Largest, Reverse ? L"(reversed)" : L"" );

	_SizeRange *SizeRange = (_SizeRange*) malloc( sizeof( _SizeRange ) );

	if ( SizeRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can't allocate size range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", SizeRange );

	SizeRange->Next		= *outPointer;
	SizeRange->Type		= 'S';
	SizeRange->Reverse	= Reverse;
	SizeRange->Smallest	= Smallest;
	SizeRange->Largest	= Largest;

	*outPointer = SizeRange;
	return 0;
}


int IsTimeSep( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '.' ) );
}


int ParseTime2( wchar_t **ch, unsigned int *Value, BOOL *HasSeconds )
{
	if ( ( ch == NULL ) || ( Value == NULL ) )
		return -1;
	else if ( iswdigit( **ch ) == 0 )
		return 1;

	unsigned int Hour = 0, Minute = 0, Second = 0;

	Hour = wcstoul( *ch, ch, 10 );

	if ( Hour > 23 )
		return 1;

	if ( !IsTimeSep( **ch ) )
		return 1;

	(*ch)++;

	if ( iswdigit( **ch ) == 0 )
		return 1;

	Minute = wcstoul( *ch, ch, 10 );

	if ( Minute > 59 )
		return 1;

	if ( IsTimeSep( **ch ) ) {
		(*ch)++;

		if ( iswdigit( **ch ) == 0 )
			return 1;

		Second = wcstoul( *ch, ch, 10 );

		if ( Second > 59 )
			return 1;

		if ( HasSeconds != NULL )
			*HasSeconds = TRUE;
	}
	else if ( HasSeconds != NULL )
		*HasSeconds = FALSE;

	if ( ( towupper( **ch ) == 'A' ) || ( towupper( **ch ) == 'P' ) ) {
		if ( ( Hour == 0 ) || ( Hour > 12 ) )
			return 1;

		if ( towupper( **ch ) == 'A' )
			Hour %= 12;
		else
			Hour = ( Hour % 12 ) + 12;

		(*ch)++;
		if ( towupper( **ch ) == 'M' )
			(*ch)++;
	}

	*Value = Second + ( 60 * Minute ) + ( 60 * 60 * Hour );
	return 0;
}


int ParseTimeRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t *sep1 = inString;
	unsigned int EarliestTime = 0, LatestTime = 0, Offset = 0;
	unsigned short int WhichStamp = TIME_STAMP_WRITE;
	BOOL EarliestHasSeconds = FALSE, LatestHasSeconds = FALSE, Add = FALSE;

	SYSTEMTIME Now;
	GetLocalTime( &Now );


	if ( towupper( *sep1 ) == 'W' ) {
		WhichStamp = TIME_STAMP_WRITE;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'A' ) {
		WhichStamp = TIME_STAMP_LAST_ACCESS;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'C' ) {
		WhichStamp = TIME_STAMP_CREATION;
		sep1++;
	}

	if ( IsColonOrEquals( *sep1 ) )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseTime2( &sep1, &EarliestTime, &EarliestHasSeconds ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad time range: \"%s\"\r\n", inString );
			return 1;
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );
		EarliestTime = ( 60 * Now.wMinute ) + ( 60 * 60 * Now.wHour );

		if ( Add ) {
			EarliestTime += ( 60 * Offset );
			if ( EarliestTime >= ( 24 * 60 * 60 ) )
				EarliestTime %= ( 24 * 60 * 60 );
		}
		else {
			if ( EarliestTime > Offset )
				EarliestTime -= Offset;
			else
				EarliestTime = 0;
		}
	}
	else if ( ( *sep1 == ',' ) || ( *sep1 == ']' ) )
		EarliestTime = ( 60 * Now.wMinute ) + ( 60 * 60 * Now.wHour );

	if ( *sep1 == ',' )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseTime2( &sep1, &LatestTime, &LatestHasSeconds ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad time range: \"%s\"\r\n", inString );
			return 1;
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );

		if ( Add ) {
			LatestTime = EarliestTime + ( 60 * Offset );
			if ( LatestTime >= ( 24 * 60 * 60 ) )
				LatestTime %= ( 24 * 60 * 60 );
		}
		else {
			if ( EarliestTime > Offset )
				LatestTime = EarliestTime - Offset;
			else
				LatestTime = 0;
		}
	}
	else if ( *sep1 == ']' )
		LatestTime = ( 60 * Now.wMinute ) + ( 60 * 60 * Now.wHour );

	if ( EarliestTime > LatestTime ) {
		unsigned int t = EarliestTime;
		EarliestTime = LatestTime;
		LatestTime = t;
		BOOL s = EarliestHasSeconds;
		EarliestHasSeconds = LatestHasSeconds;
		LatestHasSeconds = s;
	}

	if ( !LatestHasSeconds )
		LatestTime += 59;

	if ( wcscmp( sep1, L"]" ) ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad time range: \"%s\"\r\n", inString );
		return 1;
	}

	// Printf( L"   * Time range = %u - %u %s\r\n", EarliestTime, LatestTime, Reverse ? L"(reversed)" : L"" );

	_TimeRange *TimeRange = (_TimeRange*) malloc( sizeof( _TimeRange ) );

	if ( TimeRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can't allocate time range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", TimeRange );

	TimeRange->Next			= *outPointer;
	TimeRange->Type			= 'T';
	TimeRange->Reverse		= Reverse;
	TimeRange->WhichStamp	= WhichStamp;
	TimeRange->EarliestTime	= EarliestTime;
	TimeRange->LatestTime	= LatestTime;

	*outPointer = TimeRange;
	return 0;
}


BOOL Leapness( int Year )
{
	if ( ( Year % 400 ) == 0 ) return TRUE;
	if ( ( Year % 100 ) == 0 ) return FALSE;
	return ( ( Year % 4 ) == 0 );
}


unsigned int Julian( unsigned int Year, unsigned int Month, unsigned int Day )
{
	return ( ( Year - EPOCH ) * 365 ) + ( ( Year - EPOCH ) / 4 ) - ( ( Year - EPOCH ) / 100 ) + ( ( Year - EPOCH ) / 400 ) + Kalends[Month][Leapness( Year )] + Day;
}


int IsDateSep( wchar_t ch )
{
	return ( ( ch == '/' ) || ( ch == '-' ) || ( ch == '.' ) );
}


unsigned int GetLocalDateFormat()
{
	wchar_t temp[SHORT_BUF_LEN] = L"";
	unsigned int n = 0xffff;

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_IDATE, temp, SHORT_BUF_LEN ) == 0 )
		return DATE_FORMAT_MDY;

	n = wcstoul( temp, NULL, 10 );

	if ( n < 3 )
		return n;
	else
		return DATE_FORMAT_MDY;
}


int ParseDate2( wchar_t **ch, unsigned int *Value )
{
	if ( ( ch == NULL ) || ( Value == NULL ) )
		return -1;
	else if ( iswdigit( **ch ) == 0 )
		return 1;

	unsigned int Value1 = 0, Value2 = 0, Value3 = 0;
	unsigned int Year = 0, Month = 0, Day = 0;
	unsigned int DateFormat = DATE_FORMAT_MDY;

	Value1 = wcstoul( *ch, ch, 10 );

	if ( !IsDateSep( **ch ) )
		return 1;

	(*ch)++;

	if ( iswdigit( **ch ) == 0 )
		return 1;

	Value2 = wcstoul( *ch, ch, 10 );

	if ( !IsDateSep( **ch ) )
		return 1;

	(*ch)++;

	if ( iswdigit( **ch ) == 0 )
		return 1;

	Value3 = wcstoul( *ch, ch, 10 );

	if ( ( Value1 >= EPOCH ) && ( Value1 <= ULTIMA ) )
		DateFormat = DATE_FORMAT_YMD;
	else
		DateFormat = GetLocalDateFormat();

	switch ( DateFormat ) {
		case DATE_FORMAT_MDY :
			Year	= Value3;
			Month	= Value1;
			Day		= Value2;
			break;

		case DATE_FORMAT_DMY :
			Year	= Value3;
			Month	= Value2;
			Day		= Value1;
			break;

		case DATE_FORMAT_YMD :
			Year	= Value1;
			Month	= Value2;
			Day		= Value3;
			break;
	}

	if ( ( Year >= 80 ) && ( Year <= 99 ) )
		Year += 1900;
	else if ( Year <= 99 )
		Year += 2000;

	if ( ( Year < EPOCH ) || ( Year > ULTIMA ) )
		return 1;
	else if ( ( Month < 1 ) || ( Month > 12 ) )
		return 1;
	else if ( ( Day < 1 ) || ( Day > MonthLength[Month][Leapness( Year )] ) )
		return 1;

	*Value = Julian( Year, Month, Day );
	return 0;
}


int ParseDateRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t *sep1 = inString;
	unsigned int EarliestJulian = 0, LatestJulian = 0, Offset = 0;
	unsigned int EarliestTime = UINT_MAX, LatestTime = UINT_MAX;
	unsigned short int WhichStamp = TIME_STAMP_WRITE;
	BOOL EarliestHasSeconds = FALSE, LatestHasSeconds = FALSE, Add = FALSE;

	SYSTEMTIME Now;
	GetLocalTime( &Now );


	if ( towupper( *sep1 ) == 'W' ) {
		WhichStamp = TIME_STAMP_WRITE;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'A' ) {
		WhichStamp = TIME_STAMP_LAST_ACCESS;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'C' ) {
		WhichStamp = TIME_STAMP_CREATION;
		sep1++;
	}

	if ( IsColonOrEquals( *sep1 ) )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseDate2( &sep1, &EarliestJulian ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad date range: \"%s\"\r\n", inString );
			return 1;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &EarliestTime, &EarliestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad date range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );
		EarliestJulian = Julian( Now.wYear, Now.wMonth, Now.wDay );

		if ( Add )
			EarliestJulian += Offset;
		else {
			if ( EarliestJulian > Offset )
				EarliestJulian -= Offset;
			else
				EarliestJulian = 0;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &EarliestTime, &EarliestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad date range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( ( *sep1 == ',' ) || ( *sep1 == ']' ) )
		EarliestJulian = Julian( Now.wYear, Now.wMonth, Now.wDay );

	if ( *sep1 == ',' )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseDate2( &sep1, &LatestJulian ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad date range: \"%s\"\r\n", inString );
			return 1;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &LatestTime, &LatestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad date range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );

		if ( Add )
			LatestJulian = EarliestJulian + Offset;
		else {
			if ( EarliestJulian > Offset )
				LatestJulian = EarliestJulian - Offset;
			else
				LatestJulian = 0;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &LatestTime, &LatestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad date range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( *sep1 == ']' )
		LatestJulian = Julian( Now.wYear, Now.wMonth, Now.wDay );

	if ( ( EarliestJulian > LatestJulian ) || ( ( EarliestJulian == LatestJulian ) && ( EarliestTime > LatestTime ) ) ) {
		unsigned int t = EarliestJulian;
		EarliestJulian = LatestJulian;
		LatestJulian = t;
		BOOL s = EarliestHasSeconds;
		EarliestHasSeconds = LatestHasSeconds;
		LatestHasSeconds = s;
	}

	if ( EarliestTime == UINT_MAX )
		EarliestTime = 0;

	if ( ( LatestTime != UINT_MAX ) && !LatestHasSeconds )
		LatestTime += 59;

	if ( wcscmp( sep1, L"]" ) ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad date range: \"%s\"\r\n", inString );
		return 1;
	}

	// Printf( L"   * Date range = %u - %u %s\r\n", EarliestJulian, LatestJulian, Reverse ? L"(reversed)" : L"" );

	_DateRange *DateRange = (_DateRange*) malloc( sizeof( _DateRange ) );

	if ( DateRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can't allocate date range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DateRange );

	DateRange->Next				= *outPointer;
	DateRange->Type				= 'D';
	DateRange->Reverse			= Reverse;
	DateRange->WhichStamp		= WhichStamp;
	DateRange->EarliestJulian	= EarliestJulian;
	DateRange->LatestJulian		= LatestJulian;
	DateRange->EarliestTime		= EarliestTime;
	DateRange->LatestTime		= LatestTime;

	*outPointer = DateRange;
	return 0;
}


int AddExclusionRange( LPTSTR Filespec, void **outPointer )
{
	_ExclusionRange *ExclusionRange = (_ExclusionRange*) malloc( sizeof( _ExclusionRange ) + ( wcslen( Filespec ) * sizeof( wchar_t ) ) );

	if ( ExclusionRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can't allocate exclusion range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DateRange );

	ExclusionRange->Next		= *outPointer;
	ExclusionRange->Type		= '!';
	wcscpy_s( &(ExclusionRange->Filespec), wcslen( Filespec ) + 1, Filespec );

	*outPointer = ExclusionRange;
	return 0;
}


int ParseExclusionRange( LPTSTR inString, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;
	
	wchar_t Filespec[MAX_ARG_LEN] = L"";
	int j = 0;
	wchar_t ch = 0x00;
	BOOL quote = FALSE;

	for ( int i = 0; inString[i]; i++ ) {
		ch = inString[i];

		if ( iswspace( ch ) && ( j == 0 ) )
			continue;

		if ( ch == '"' ) {
			quote = !quote;
			continue;
		}

		if ( !quote && iswspace( ch ) ) {
			if ( TrimString( Filespec ) > 0 ) {
				AddExclusionRange( Filespec, outPointer );
				j = 0;
				Filespec[j] = 0x00;
			}
		}
		
		if ( ( ch == ']' ) && ( inString[i + 1] == 0x00 ) ) {
			if ( TrimString( Filespec ) > 0 ) {
				AddExclusionRange( Filespec, outPointer );
				return 0;
			}
		}

		if ( j < ( MAX_ARG_LEN - 1 ) ) {
			Filespec[j++] = ch;
			Filespec[j] = 0x00;
		}
	}

	return 0;
}


int ParseDOW( LPTSTR inString )
{
	if ( inString == NULL )
		return -1;

	int num = -1;
	wchar_t *sep1 = '\0';

	if ( iswdigit( inString[0] ) ) {
		num = wcstol( inString, &sep1, 10 );
		if ( *sep1 != '\0' )
			return -1;

		if ( ( num >= 0 ) && ( num <= 7 ) )
			return num % 7;
		else
			return -1;
	}

	for ( int i = 0; i < 7; i++ )
		if ( _wcsnicmp( inString, DowName[i], 2 ) == 0 )
			return i;

	return -1;
}


int ParseDOW2( LPTSTR inString, int *len )
{
	if ( inString == NULL )
		return -1;

	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	int i = 0;
	wchar_t ch = '\0';

	do {
		ch = inString[i];
		if ( iswspace( ch ) || ( ch == ';' ) || ( ch == '-' ) || ( ch == '.' ) || ( ch == ',' ) || ( ch == ']' ) || ( i >= ( SHORT_BUF_LEN - 1 ) ) )
			ch = '\0';

		shortbuf[i++] = ch;

	} while ( ch );

	if ( len )
		*len = i - 1;

	if ( ( _wcsicmp( shortbuf, L"weekday" ) == 0 ) || ( _wcsicmp( shortbuf, L"weekdays" ) == 0 ) )
		return WEEKDAYS_MASK;
	else if ( ( _wcsicmp( shortbuf, L"weekend" ) == 0 ) || ( _wcsicmp( shortbuf, L"weekends" ) == 0 ) )
		return WEEKENDS_MASK;
	else
		return ParseDOW( shortbuf );
}


int ParseDOWRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t *ch = inString;
	int i = 0;
	unsigned short int DOWMask = 0;
	unsigned short int WhichStamp = TIME_STAMP_WRITE;
	int dow = 0, dow2 = 0;
	int len = 0;

	const wchar_t ErrorStr[] = L"Bad day-of-the-week range: \"%s\"\r\n";

	
	if ( towupper( inString[i] ) == 'W' ) {
		WhichStamp = TIME_STAMP_WRITE;
		i++;
	}
	else if ( towupper( inString[i] ) == 'A' ) {
		WhichStamp = TIME_STAMP_LAST_ACCESS;
		i++;
	}
	else if ( towupper( inString[i] ) == 'C' ) {
		WhichStamp = TIME_STAMP_CREATION;
		i++;
	}

	if ( IsColonOrEquals( inString[i] ) )
		i++;
	else {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), ErrorStr, inString );
		return 1;
	}


	while ( inString[i] ) {
		dow = ParseDOW2( inString + i, &len );
		if ( dow < 0 ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), ErrorStr, inString );
			return 1;
		}
		else
			i += len;

		dow2 = -1;
		if ( ( i < ( MAX_ARG_LEN - 4 ) ) && ( dow <= 6 ) ) {
			if ( ( inString[i] == '-' ) || ( ( inString[i] == '.' ) && ( inString[i + 1] == '.' ) ) ) {
				// parse dow range:

				i += ( inString[i] == '.' ) ? 2 : 1;

				dow2 = ParseDOW2( inString + i, &len );
				if ( ( dow2 < 0 ) || ( dow2 > 6 ) ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), ErrorStr, inString );
					return 1;
				}
				else
					i += len;

				if ( ( inString[i] == '-' ) || ( inString[i] == '.' ) ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), ErrorStr, inString );
					return 1;
				}
			}
		}

		while ( iswspace( inString[i] ) || ( inString[i] == ',' ) || ( inString[i] == ';' ) )
			i++;

		if ( dow <= 6 )
			DOWMask |= ( 1 << dow );
		else
			DOWMask |= dow;

		while ( ( dow2 >= 0 ) && ( dow != dow2 ) ) {
			dow = ( ( dow + 1 ) % 7 );
			DOWMask |= ( 1 << dow );
		}

		if ( wcscmp( inString + i, L"]" ) == 0 )
			break;
	}

	if ( Reverse )
		DOWMask ^= 0x7f;

	if ( DOWMask == 0 ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Empty day-of-the-week range: \"%s\"\r\n", inString );
		return 1;
	}


	// Printf( L"   * DOW mask = %u\r\n", DOWMask );

	_DOWRange *DOWRange = (_DOWRange*) malloc( sizeof( _DOWRange ) );

	if ( DOWRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can't allocate DOW range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DOWRange );

	DOWRange->Next			= *outPointer;
	DOWRange->Type			= 'W';
	DOWRange->WhichStamp	= WhichStamp;
	DOWRange->DOWMask		= DOWMask;

	*outPointer = DOWRange;
	return 0;
}


int ParseDescriptionRange( LPTSTR inString, BOOL Reverse, BOOL SlashI, void **outPointer )
{
	if ( inString == NULL )
		return -1;
	else if ( inString[0] == 0x00 )
		return 0;
	else if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t Description[MAX_ARG_LEN] = L"";
	int i = 0, len = 0;
	
	if ( IsColonOrEquals( inString[i] ) )
		i++;

	wcscpy_s( Description, MAX_ARG_LEN, inString + i );

	if ( !SlashI ) {
		len = (int) wcslen( Description );

		if ( len )
			if ( Description[len - 1] == ']' )
				Description[len - 1] = 0x00;
	}

	// Printf( L"   * Description range = '%s' %s\r\n", Description, Reverse ? L"(reversed)" : L"" );

	_DescriptionRange *DescriptionRange = (_DescriptionRange*) malloc( sizeof( _DescriptionRange ) + ( wcslen( Description ) * sizeof( wchar_t ) ) );

	if ( DescriptionRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can't allocate description range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DescriptionRange );

	DescriptionRange->Next			= *outPointer;
	DescriptionRange->Type			= 'I';
	DescriptionRange->Reverse		= Reverse;
	wcscpy_s( &(DescriptionRange->Description), wcslen( Description ) + 1, Description );

	*outPointer = DescriptionRange;
	return 0;
}


int ParseOwnerRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( inString == NULL )
		return -1;
	else if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t Owner[MAX_ARG_LEN] = L"";
	int i = 0, len = 0;
	
	if ( IsColonOrEquals( inString[i] ) )
		i++;

	wcscpy_s( Owner, MAX_ARG_LEN, inString + i );

	len = (int) wcslen( Owner );

	if ( len )
		if ( Owner[len - 1] == ']' )
			Owner[len - 1] = 0x00;

	// Printf( L"   * Owner range = '%s' %s\r\n", Owner, Reverse ? L"(reversed)" : L"" );

	_OwnerRange *OwnerRange = (_OwnerRange*) malloc( sizeof( _OwnerRange ) + ( wcslen( Owner ) * sizeof( wchar_t ) ) );

	if ( OwnerRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can't allocate owner range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", OwnerRange );

	OwnerRange->Next			= *outPointer;
	OwnerRange->Type			= 'O';
	OwnerRange->Reverse			= Reverse;
	wcscpy_s( &(OwnerRange->Owner), wcslen( Owner ) + 1, Owner );

	*outPointer = OwnerRange;
	return 0;
}


int ParseRangeOption( LPTSTR inString, void **outPointer )
{
	if ( ( inString == NULL ) || ( outPointer == NULL ) )
		return -1;
	else if ( inString[0] == 0x00 )
		return -1;

	int j = 0;
	BOOL Reverse = FALSE;

	if ( inString[j] != '/' )
		return 1;

	j++;

	if ( inString[j] == '!' ) {
		Reverse = TRUE;
		j++;
	}

	if ( inString[j] != '[' )
		return 1;

	j++;

	if ( inString[wcslen(inString) - 1] != ']' ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Range option must end with ']': \"%s\"\r\n", inString );
		return 1;
	}

	switch ( towupper( inString[j++] ) ) {

		case 'S' :
			return ParseSizeRange( inString + j, Reverse, outPointer );

		case 'T' :
			return ParseTimeRange( inString + j, Reverse, outPointer );

		case 'D' :
			return ParseDateRange( inString + j, Reverse, outPointer );

		case '!' :
			return ParseExclusionRange( inString + j, outPointer );

		case 'W' :
			return ParseDOWRange( inString + j, Reverse, outPointer );

		case 'I' :
			return ParseDescriptionRange( inString + j, Reverse, FALSE, outPointer );

		case 'O' :
			return ParseOwnerRange( inString + j, Reverse, outPointer );

		default :
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown range type: \"%s\"\r\n", inString );
			return 1;
	}

	return 0;
}


int ParseAttributesOption( LPTSTR inString, BOOL AllowDirectories, DWORD *AttribMask, DWORD *AttribValues )
{
	if ( inString == NULL )
		return -1;
	else if ( inString[0] == 0x00 )
		return -1;

	int j = 0;
	if ( inString[j] == '/' )
		j++;

	if ( iswalpha( inString[j] ) )
		j++;
	
	if ( IsColonOrEquals( inString[j] ) )
		j++;

	if ( inString[j] == 0x00 ) {
		*AttribMask = 0;
		*AttribValues = 0;
		return 0;
	}


	BOOL Minus = FALSE;
	const wchar_t ErrorString[] = L"Bad attributes: \"%s\"\r\n";


	while ( inString[j] ) {

		switch ( towupper( inString[j] ) ) {

			case 'A' :
				*AttribMask |= FILE_ATTRIBUTE_ARCHIVE;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_ARCHIVE;
				else
					*AttribValues |= FILE_ATTRIBUTE_ARCHIVE;

				Minus = FALSE;
				break;

			case 'C' :
				*AttribMask |= FILE_ATTRIBUTE_COMPRESSED;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_COMPRESSED;
				else
					*AttribValues |= FILE_ATTRIBUTE_COMPRESSED;

				Minus = FALSE;
				break;

			case 'D' :
				if ( !AllowDirectories ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Directory bit not allowed: \"%s\"\r\n", inString );
					return 1;
				}

				*AttribMask |= FILE_ATTRIBUTE_DIRECTORY;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_DIRECTORY;
				else
					*AttribValues |= FILE_ATTRIBUTE_DIRECTORY;

				Minus = FALSE;
				break;

			case 'E' :
				*AttribMask |= FILE_ATTRIBUTE_ENCRYPTED;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_ENCRYPTED;
				else
					*AttribValues |= FILE_ATTRIBUTE_ENCRYPTED;

				Minus = FALSE;
				break;

			case 'H' :
				*AttribMask |= FILE_ATTRIBUTE_HIDDEN;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_HIDDEN;
				else
					*AttribValues |= FILE_ATTRIBUTE_HIDDEN;

				Minus = FALSE;
				break;

			case 'I' :
				*AttribMask |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;
				else
					*AttribValues |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;

				Minus = FALSE;
				break;

			case 'O' :
				*AttribMask |= FILE_ATTRIBUTE_OFFLINE;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_OFFLINE;
				else
					*AttribValues |= FILE_ATTRIBUTE_OFFLINE;

				Minus = FALSE;
				break;

			case 'R' :
				*AttribMask |= FILE_ATTRIBUTE_READONLY;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_READONLY;
				else
					*AttribValues |= FILE_ATTRIBUTE_READONLY;

				Minus = FALSE;
				break;

			case 'S' :
				*AttribMask |= FILE_ATTRIBUTE_SYSTEM;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_SYSTEM;
				else
					*AttribValues |= FILE_ATTRIBUTE_SYSTEM;

				Minus = FALSE;
				break;

			case '-' :
				if ( iswalpha( inString[j + 1] ) == 0 ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), ErrorString, inString );
					return 1;
				}

				Minus = TRUE;
				break;

			default :
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), ErrorString, inString );
				return 1;
		}

		j++;
	}

	return 0;
}



int GetFileOwner( LPTSTR Filename, LPTSTR Owner )
{
	if ( ( Filename == NULL ) || ( Owner == NULL ) )
		return -1;

	SECURITY_INFORMATION *FileSecurity = NULL;
	DWORD FileSecSize = 0;

	SID *OwnerSid = NULL;
	BOOL Dummy;

	wchar_t OwnerName[MAX_ARG_LEN]		= L"";
	wchar_t OwnerDomain[MAX_ARG_LEN]	= L"";
	DWORD OwnerNameSize = MAX_ARG_LEN, OwnerDomainSize = MAX_ARG_LEN;

	SID_NAME_USE SidNameUse = SidTypeUnknown;

	int rv = 1;


	GetFileSecurity( Filename, OWNER_SECURITY_INFORMATION, NULL, 0, &FileSecSize );
	if ( FileSecSize == 0 )
		return 1;

	FileSecurity = (SECURITY_INFORMATION*) malloc( FileSecSize );

	if ( FileSecurity ) {
		if ( GetFileSecurity( Filename, OWNER_SECURITY_INFORMATION, FileSecurity, FileSecSize, &FileSecSize ) )
			if ( GetSecurityDescriptorOwner( FileSecurity, (PSID*) &OwnerSid, &Dummy ) )
				if ( LookupAccountSid( NULL, OwnerSid, OwnerName, &OwnerNameSize, OwnerDomain, &OwnerDomainSize, &SidNameUse ) )
					rv = 0;

		free( FileSecurity );
	}

	if ( OwnerName[0] && OwnerDomain[0] ) {
		wcscpy_s( Owner, MAX_ARG_LEN, OwnerDomain );
		wcscat_s( Owner, MAX_ARG_LEN, L"\\" );
		wcscat_s( Owner, MAX_ARG_LEN, OwnerName );
	}

	return rv;
}


int CheckRanges( WIN32_FIND_DATA *FoundItem, LPTSTR PathName, void *FirstRange )
{
	if ( ( FoundItem == NULL ) || ( PathName == NULL ) )
		return -1;

	if ( FirstRange == NULL )
		return 0;


	_SizeRange *SizeRange					= (_SizeRange*) FirstRange;
	_TimeRange *TimeRange					= NULL;
	_DateRange *DateRange					= NULL;
	_ExclusionRange *ExclusionRange			= NULL;
	_DOWRange *DOWRange						= NULL;
	_DescriptionRange *DescriptionRange		= NULL;
	_OwnerRange *OwnerRange					= NULL;

	unsigned __int64 FileSize = 0;
	FILETIME *FileStamp = NULL;
	FILETIME LocalFileTime;
	SYSTEMTIME FileTimeUnpacked;
	unsigned int FileTimeAsSeconds = 0;
	unsigned int FileDateAsJulian = 0;
	wchar_t FullName[MAX_FILENAME_LEN]	= L"";
	wchar_t StringBuffer[MAX_ARG_LEN]	= L"";


	BOOL Okay = FALSE;


	while ( SizeRange ) {

		switch ( towupper( SizeRange->Type ) ) {

			case 'S' :
				FileSize = ( (__int64) FoundItem->nFileSizeHigh << 32 ) | FoundItem->nFileSizeLow;

				Okay = ( ( FileSize >= SizeRange->Smallest ) && ( FileSize <= SizeRange->Largest ) );
				if ( SizeRange->Reverse )
					Okay = !Okay;

				break;

			case 'T' :
				TimeRange = (_TimeRange*) SizeRange;

				if ( TimeRange->WhichStamp == TIME_STAMP_LAST_ACCESS )
					FileStamp = &FoundItem->ftLastAccessTime;
				else if ( TimeRange->WhichStamp == TIME_STAMP_CREATION )
					FileStamp = &FoundItem->ftCreationTime;
				else
					FileStamp = &FoundItem->ftLastWriteTime;

				FileTimeToLocalFileTime( FileStamp, &LocalFileTime );
				FileTimeToSystemTime( &LocalFileTime, &FileTimeUnpacked );
				FileTimeAsSeconds = FileTimeUnpacked.wSecond + ( 60 * FileTimeUnpacked.wMinute ) + ( 60 * 60 * FileTimeUnpacked.wHour );

				Okay = ( ( FileTimeAsSeconds >= TimeRange->EarliestTime ) && ( FileTimeAsSeconds <= TimeRange->LatestTime ) );
				if ( TimeRange->Reverse )
					Okay = !Okay;

				break;

			case 'D' :
				DateRange = (_DateRange*) SizeRange;

				if ( DateRange->WhichStamp == TIME_STAMP_LAST_ACCESS )
					FileStamp = &FoundItem->ftLastAccessTime;
				else if ( DateRange->WhichStamp == TIME_STAMP_CREATION )
					FileStamp = &FoundItem->ftCreationTime;
				else
					FileStamp = &FoundItem->ftLastWriteTime;

				FileTimeToLocalFileTime( FileStamp, &LocalFileTime );
				FileTimeToSystemTime( &LocalFileTime, &FileTimeUnpacked );
				FileDateAsJulian = Julian( FileTimeUnpacked.wYear, FileTimeUnpacked.wMonth, FileTimeUnpacked.wDay );
				FileTimeAsSeconds = FileTimeUnpacked.wSecond + ( 60 * FileTimeUnpacked.wMinute ) + ( 60 * 60 * FileTimeUnpacked.wHour );

				Okay = ( ( FileDateAsJulian > DateRange->EarliestJulian ) || ( ( FileDateAsJulian == DateRange->EarliestJulian ) && ( FileTimeAsSeconds >= DateRange->EarliestTime ) ) );
				Okay &= ( ( FileDateAsJulian < DateRange->LatestJulian ) || ( ( FileDateAsJulian == DateRange->LatestJulian ) && ( FileTimeAsSeconds <= DateRange->LatestTime ) ) );
				if ( DateRange->Reverse )
					Okay = !Okay;

				break;

			case '!' :
				ExclusionRange = (_ExclusionRange*) SizeRange;
				Okay = ( WildcardComparison( &(ExclusionRange->Filespec), FoundItem->cFileName, 1, 1 ) != 0 );
				break;

			case 'W' :
				DOWRange = (_DOWRange*) SizeRange;

				if ( DOWRange->WhichStamp == TIME_STAMP_LAST_ACCESS )
					FileStamp = &FoundItem->ftLastAccessTime;
				else if ( DOWRange->WhichStamp == TIME_STAMP_CREATION )
					FileStamp = &FoundItem->ftCreationTime;
				else
					FileStamp = &FoundItem->ftLastWriteTime;

				FileTimeToLocalFileTime( FileStamp, &LocalFileTime );
				FileTimeToSystemTime( &LocalFileTime, &FileTimeUnpacked );
				FileDateAsJulian = Julian( FileTimeUnpacked.wYear, FileTimeUnpacked.wMonth, FileTimeUnpacked.wDay );

				Okay = DOWRange->DOWMask & ( 1 << ( FileDateAsJulian % 7 ) );
				break;

			case 'I' :
				DescriptionRange = (_DescriptionRange*) SizeRange;

				wcscpy_s( FullName, MAX_FILENAME_LEN, PathName );
				_Ranges_AppendFilename( FullName, FoundItem->cFileName, MAX_FILENAME_LEN );		//	MakeDirectoryName( FullName, FoundItem->cFileName );
				StringBuffer[0] = 0x00;
				_Ranges_GetFileDescription( FullName, StringBuffer, MAX_ARG_LEN );				//	ReadDescriptions( FullName, StringBuffer );

				Okay = ( WildcardComparison( &(DescriptionRange->Description), StringBuffer, 0, 1 ) == 0 );
				if ( DescriptionRange->Reverse )
					Okay = !Okay;

				break;

			case 'O' :
				OwnerRange = (_OwnerRange*) SizeRange;

				wcscpy_s( FullName, MAX_FILENAME_LEN, PathName );
				_Ranges_AppendFilename( FullName, FoundItem->cFileName, MAX_FILENAME_LEN );		//	MakeDirectoryName( FullName, FoundItem->cFileName );
				StringBuffer[0] = 0x00;

				GetFileOwner( FullName, StringBuffer );

				Okay = ( WildcardComparison( &(OwnerRange->Owner), StringBuffer, 0, 1 ) == 0 );
				if ( OwnerRange->Reverse )
					Okay = !Okay;

				break;
		}

		if ( !Okay )
			return 1;
		
		SizeRange = (_SizeRange*) SizeRange->Next;
	}


	return 0;
}


